# -*- coding: utf-8 -*-
"""
Copyright © 2017 United States Government as represented by the Administrator, National Aeronautics and Space Administration.  
No Copyright is claimed in the United States under Title 17, U.S. Code.  All Other Rights Reserved.
"""
import logging
import os
import tempfile
import sys
import numpy.f2py  # just to check it presents
from numpy.distutils.exec_command import exec_command


import sympy
from collections import OrderedDict, defaultdict
import collections
import functools
from sympy import sympify as parse_expr
from sympy.utilities.autowrap import ufuncify
import functools
from decorator import decorator, decorate
from sympy import symbols, Symbol
from sympy.core.function import UndefinedFunction
from inspect import getargspec, getfullargspec
import inspect
from sympy.physics import units
from sympy.physics import units as sympy_units
from sympy.physics.units.systems.si import dimsys_SI
import numpy as np
from sympy import latex, Eq
from sympy.parsing.latex import parse_latex
import pandas as pd
from datetime import datetime
from scipy.integrate import solve_ivp
from sympy.physics.units.util import _get_conversion_matrix_for_expr

from sympy.core.compatibility import reduce, Iterable, ordered
from sympy import Add, Mul, Pow, Tuple, sympify, default_sort_key
from sympy.physics.units.quantities import Quantity
from sympy.physics.units import Dimension
from sympy import nsimplify
from sympy import Function

import urllib.request, json
import requests

import base64
import types
import forge

from sympy.physics.units import UnitSystem
from sympy.physics.units import Dimension




def get_unit_quantity(name, base, scale_factor, abbrev=None, unit_system='SI'):
    '''Define a unit in terms of a base unit'''
    unit = units.Quantity(name, abbrev=abbrev)
    base_unit = getattr(sympy_units, base)

    try:
        # sympy >= 1.5 raises the following warning:
        #   Use unit_system.set_quantity_dimension or
        # <unit>.set_global_relative_scale_factor
        unit.set_global_relative_scale_factor(scale_factor, base_unit)
    except AttributeError:
        unit.set_dimension(base_unit.dimension)
        unit.set_scale_factor(scale_factor * base_unit, unit_system=unit_system)

    return unit


unit_subs = dict(
    nT=get_unit_quantity('nanotesla', 'tesla', .000000001, 'nT', 'SI'),
    R_E=get_unit_quantity('earth radii', 'm', 6.371e6, 'R_E', 'SI'),
    R_S=get_unit_quantity('solar radii', 'm', 6.957e8, 'R_S', 'SI'),
    erg=get_unit_quantity('erg', 'J', .0000001, 'erg', 'SI'),
    nPa=get_unit_quantity('nanopascals', 'pascal', .000000001, 'nPa', 'SI'),
    cc=sympy_units.cm**3,
    AU=get_unit_quantity('astronomical unit', 'm', 1.496e+11, 'AU', 'SI'),
    arcsec=get_unit_quantity('arc seconds', 'degrees', 1./3600, '\"', 'SI'),
    hr = get_unit_quantity('hour','s',3600.,'hr','SI')
    #TECU = get_unit_quantity('TECU','1/m**2',10**16, 'TECU','SI')
    )

sympy_units.erg = unit_subs['erg']

prefix_dict = sympy_units.prefixes.PREFIXES  #built-in dictionary of Prefix instances
#test_unit_subs={}  #dictionary to replace subs in kamodo.get_unit_quantities()
unit_list = ['m', 's', 'g', 'A', 'K', 'radian', 'sr', 'cd', 'mole', 'eV', 'Pa', 'F', 'N',
             'V', 'Hz', 'C', 'W', 'Wb', 'H', 'S', 'Bq', 'Gy', 'erg', 'T']

# list of SI units included in sympy (likely not complete)

for item in unit_list:
    unit_item = getattr(sympy_units, item)
    unit_subs[item] = unit_item
    for key in prefix_dict.keys():
        unit_subs[key+item] = get_unit_quantity(str(prefix_dict[key].name)+
                                                str(unit_item.name), str(unit_item.name),
                                                float(prefix_dict[key].scale_factor),
                                                abbrev=key+item)

reserved_names = dir(sympy)


def get_kamodo_unit_system():
    """Same as SI but supports anglular frequency"""

    radian = sympy.physics.units.radian
    degree = sympy.physics.units.degree
    si_unit_system = UnitSystem.get_unit_system('SI')
    si_dimension_system = si_unit_system.get_dimension_system()

    angle = Dimension('angle', 'A')

    kamodo_dims = si_dimension_system.extend(
        new_base_dims=(angle,),
        new_derived_dims=[Dimension('angular_velocity')],
        new_dim_deps={Symbol('angular_velocity'): {Symbol('angle'): 1, Symbol('time'): -1}})

    kamodo_units = si_unit_system.extend(
        (radian,),
        (radian, degree),
        dimension_system=kamodo_dims)

    return kamodo_units


kamodo_unit_system = get_kamodo_unit_system()

def get_ufunc(expr, variable_map):
    """Numerically optimize expression"""

    expr = sympify(expr, locals=variable_map)
    func = ufuncify(expr.free_symbols, expr)
    formula = 'f{} = {}'.format(tuple(expr.free_symbols), expr)
    return func, formula


def compile_fortran(source, module_name, extra_args='', folder='./'):
    """compile fortran source code"""
    with tempfile.NamedTemporaryFile('w', suffix='.f90') as fortran_file:
        fortran_file.write(source)
        fortran_file.flush()

        args = ' -c -m {} {} {}'.format(module_name, fortran_file.name, extra_args)
        command = 'cd "{}" && "{}" -c "import numpy.f2py as f2py;f2py.main()" {}'.format(
            folder, sys.executable, args)
        status, output = exec_command(command)
        return status, output, command


def substr_replace(name, name_maps):
    """replaces all substrings in name with those given by name_maps"""
    for old, new in name_maps:
        name = name.replace(old, new)
    return name


# This should be in yaml
def beautify_latex(expr):
    """convert string to latex-compatible expression"""
    return substr_replace(expr, [
        ('**', '^'),
        ('plus', '+'),
        ('minus', '-'),
        ('comma', ','),
        ('LEFT', '\\left ('),
        ('RIGHT', '\\right )'),
        ('integral', '\\int'),
        ('rvert', '\\rvert'),
        ('lvert', '\\lvert'),
    ])


def arg_to_latex(arg):
    return beautify_latex(latex(Symbol(arg)))


def decorator_wrapper(f, *args, **kwargs):
    """Wrapper needed by decorator.decorate to pass through args, kwargs"""
    return f(*args, **kwargs)


def kamodofy(
        _func=None,
        units='',
        arg_units=None,
        data=None,
        update=None,
        equation=None,
        citation=None,
        hidden_args=[],
        **kwargs):
    """Adds meta and data attributes to functions for compatibility with Komodo

    meta: a dictionary containing {units: <str>}
    data:
        if supplied, set f.data = data
        if not supplied, set f.data = f(), assuming it can be called with no arguments.
            If f cannot be called with no arguments, set f.data = None
    """

    def decorator_kamodofy(f):
        f.meta = dict(
            units=units,
            arg_units=arg_units,
            citation=citation,
            equation=equation,
            hidden_args=hidden_args)

        if citation is not None:
            if f.__doc__ is not None:
                f.__doc__ = f.__doc__ + '\n\ncitation: {}'.format(citation)
            else:
                f.__doc__ = 'citation: {}'.format(citation)
        f.update = update
        if data is None:
            try:
                f.data = f(**kwargs)
            except:
                f.data = None
        else:
            f.data = data

        if equation is not None:
            latex_str = equation.strip("$")
            f._repr_latex_ = lambda: latex_str
        # f._repr_latex_ = lambda : "${}$".format(latex(parse_latex(latex_str)))
        else:
            f_ = symbols(f.__name__, cls=UndefinedFunction)
            lhs = f_.__call__(*symbols(getfullargspec(f).args))
            # lambda_ = symbols('lambda', cls=UndefinedFunction)
            lambda_ = Function('lambda')
            latex_eq = latex(Eq(lhs, lambda_(*lhs.args)))
            f._repr_latex_ = lambda: "${}$".format(latex(latex_eq))

        return decorate(f, decorator_wrapper)  # preserves signature

    if _func is None:
        return decorator_kamodofy
    else:
        return decorator_kamodofy(_func)



def sort_symbols(symbols):
    symbols_ = list(symbols)
    symbols_.sort(key=str)
    return tuple(symbols_)


def valid_args(f, kwargs):
    '''Extract arguments from kwargs that appear in f'''
    valid = OrderedDict()
    if type(f) == np.vectorize:
        for a in inspect.getfullargspec(f.pyfunc).args:
            if a in kwargs:
                valid[a] = kwargs[a]
        return valid
    else:
        for a in getfullargspec(f).args:
            if a in kwargs:
                valid[a] = kwargs[a]
        return valid


def eval_func(func, kwargs):
    '''Evaluate function over valid argument'''
    try:
        return func(**valid_args(func, kwargs))
    except TypeError as m:
        raise TypeError(str(m) + str(list(valid_args(func, kwargs).keys())))


def get_defaults(func):
    sig = inspect.signature(func)
    defaults = {}
    for k, v in sig.parameters.items():
        if v.default is not inspect._empty:
            defaults[k] = v.default

    return defaults

def get_args(func):
    sig = inspect.signature(func)
    return tuple(sig.parameters.keys())

def cast_0_dim(a, to):
    if a.ndim == 0:
        return a * np.ones(to.shape)
    else:
        return a


def simulate(state_funcs, **kwargs):
    """Iterate over functions.

    state_funcs(OrderedDict)
        key: the variable to update
        value: the function that updates the variable

    Any remaining kwargs are passed to the state functions.

    The order of the keys in state_funcs determines which variables get updated first.
    """

    verbose = kwargs.get('verbose', False)
    steps = kwargs.get('steps', 1)

    state_dict = kwargs.copy()
    yield OrderedDict([(k, state_dict.get(k, None)) for k in state_funcs])
    for i in range(steps):
        result = []
        for arg, f in list(state_funcs.items()):
            try:
                result.append((arg, eval_func(f, state_dict)))
            except TypeError as m:
                raise TypeError('{}:'.format(arg) + str(m))
            state_dict.update(OrderedDict(result))
        yield OrderedDict(result)


def pad_nan(array):
    if isinstance(array, types.GeneratorType):
        return np.array(list(array))
    try:
        if len(array.shape) == 1:
            # print('padding array {}'.format(array.shape))
            return np.pad(array.astype(float), (0, 1),
                          mode='constant', constant_values=np.nan).T
        elif len(array.shape) == 2:
            # print('padding array {}'.format(array.shape))
            return np.pad(array.astype(float), ((0, 1), (0, 0)),
                          mode='constant', constant_values=np.nan)
        else:
            raise NotImplementedError('cannot pad shape {}'.format(array.shape))
    except:
        raise NotImplementedError('cannot handle {} of type {}'.format(array, type(array)))


def concat_solution(gen, variable):
    """combine solutions of a function generator
    iterates through a function generator and extracts defaults for each function
    these solutions are then padded with nans and stacked

    stacking is vertically for N-d, horizontally for 1-d
    """
    params = defaultdict(list)
    for f in gen:
        for k, v in list(get_defaults(f).items()):
            params[k].append(pad_nan(v))
        params[variable].append(pad_nan(f.data))

    for k, v in list(params.items()):
        if len(v[0].shape) == 1:
            params[k] = np.hstack(v)
        else:
            params[k] = np.vstack(v)
    return params


existing_plot_types = pd.DataFrame({
    ('1', 'N', 'N'): ['1-d line', ''],
    ('1', 'N', 'Nx2'): ['2-d line', ''],
    ('1', 'N', 'Nx3'): ['3-d line', ''],
    ('3', 'N, N, N', 'N'): ['3-d colored line', ''],
    ('1', 'Nx2', 'Nx2'): ['2-d vector field', ''],
    ('1', 'Nx3', 'Nx3'): ['3-d vector field', ''],
    ('2', 'N, M', 'NxM'): ['2-d contour', 'indexing'],
    ('2', 'NxM, NxM', 'NxM'): ['2-d contour (skew/carpet)', 'indexing'],
    ('3', 'NxM, NxM, NxM', '1'): ['Parametric Surface', ''],
    ('3', 'NxM, NxM, NxM', 'NxM'): ['Coloured Parametric Surface', ''],
    ('3', 'L, M, 1', 'LxM'): ['Map-to-plane', 'indexing*'],
    ('3', '1, M, N', 'MxN'): ['Map-to-plane', 'indexing*'],
    ('3', 'L, 1, N', 'LxN'): ['Map-to-plane', 'indexing*']
}).T
existing_plot_types.index.set_names(['nargs', 'arg shapes', 'out shape'], inplace=True)
existing_plot_types.columns = ['Plot Type', 'notes']


def gridify(_func=None, order='A', squeeze=True, **defaults):
    """Given a function of shape (n,dim) and arguments of shape (L), (M), calls f with points L*M
    
    order: 'A' (default) uses indexing='xy' in meshgrid
           'C' uses indexing='ij' in meshgrid
    squeeze: True (default) passed to reshape before returning

    """

    def decorator_gridify(f):

        signature = []
        for arg, arg_default in defaults.items():
            signature.append(forge.arg(arg, default=arg_default))

        if order == 'A':
            indexing = 'xy'
        else:
            indexing = 'ij'

        @forge.sign(*signature)
        def wrapped(**kwargs):
            coordinates = np.meshgrid(
                *kwargs.values(),
                indexing=indexing,
                sparse=False,
                copy=False)
            points = np.column_stack([c.ravel() for c in coordinates])

            if squeeze:
                out_shape = [-1] + list(coordinates[0].shape)
                return np.squeeze(f(points).reshape(out_shape, order=order))
            else:
                out_shape = list(coordinates[0].shape)
                return f(points).reshape(out_shape, order=order)

        wrapped.__name__ = f.__name__
        wrapped.__doc__ = f.__doc__

        return decorate(wrapped, decorator_wrapper)

    if _func is None:
        return decorator_gridify
    else:
        return decorator_gridify(_func)


def pointlike(_func=None, signature=None, otypes=[np.float], squeeze=None):
    """Transforms a single-argument function to one that accepts m points of dimension n"""

    def decorator_pointlike(func):
        def argument_wrapper(f, *args, **kwargs):
            """Wrapper needed by decorator.decorate to pass through args, kwargs"""
            if type(args[0]) != np.array:
                args = [np.array(x) for x in args]
            for i, x in enumerate(args):
                if len(x.shape) == 1:
                    args[i] = np.expand_dims(x, axis=0)
            if squeeze is not None:
                try:
                    return np.vectorize(f, otypes=otypes, signature=signature)(*args, **kwargs).squeeze(squeeze)
                except:
                    return np.vectorize(f, otypes=otypes, signature=signature)(*args, **kwargs)
            else:
                return np.vectorize(f, otypes=otypes, signature=signature)(*args, **kwargs)

        if not hasattr(func, '__name__'):
            func.__name__ = 'pointlike'

        return decorate(func, argument_wrapper)

    if _func is None:
        return decorator_pointlike
    else:
        return decorator_pointlike(_func)


def event(func, terminal=True, direction=0):
    def wrapped(*args, **kwargs):
        return func(*args, **kwargs)

    wrapped.terminal = terminal
    wrapped.direction = direction
    return wrapped


def solve(fprime=None, seeds=None, varname=None, interval=None,
          dense_output=True,  # generate a callable solution
          events=None,  # stop when event is triggered
          vectorized=True,
          npoints=50,
          directions=(-1, 1),
          verbose=False,
          ):
    """Decorator that solves initial value problem for a given function

    Can be used to generate streamlines, streaklines, fieldlines, etc
    """

    if len(seeds.shape) > 1:
        pass
    else:
        seeds = np.expand_dims(seeds, axis=0)

    t_eval = np.linspace(*interval, npoints)
    nseeds = len(seeds)

    def decorator_solve(f):
        solutions = []
        t = []

        fprime_ = {}
        for d in directions:
            fprime_[d] = lambda s, y: d * f(y.T)

        for i, seed in enumerate(seeds):
            for d in directions:
                result = solve_ivp(fprime_[d], interval, seed,
                                   dense_output=dense_output,
                                   events=events,
                                   vectorized=vectorized,
                                   t_eval=t_eval)
                solutions.append(result['sol'])
                interval_bounded = result['t']
                seed_numbers = np.ones(len(interval_bounded)) * i  # *len(directions) + 1*(d > 0)
                integrals = interval_bounded[::d] * d * 1j
                if d < 0:
                    t.extend(list(seed_numbers + integrals))
                else:
                    t.extend(list(seed_numbers + integrals)[1:])

        t = np.hstack(t)

        def solution(s=t):
            s = np.array(s)
            if len(s.shape) == 0:
                s = np.expand_dims(s, axis=0)

            isolution = np.floor(s.real).astype(int) * len(directions) + (s.imag > 0)

            results = []
            seed_number = []
            integral = []
            for soln, imag_ in zip(isolution, s.imag):
                seed_number.append(np.floor(soln / len(directions)))
                integral.append(imag_)
                try:
                    if np.isnan(abs(soln + imag_)):
                        results.append(np.ones(isolution.shape[-1]) * np.nan)
                    else:
                        results.append(solutions[soln](np.abs(imag_)))
                except:
                    results.append(np.ones(isolution.shape[-1]) * np.nan)
            index_ = pd.MultiIndex.from_arrays([seed_number, integral],
                                               names=['seed', 'integral'])
            return pd.DataFrame(np.vstack(results), index=index_).drop_duplicates()

        solution.__name__ = varname

        return decorate(solution, decorator_wrapper)  # preserves signature

    if fprime is None:
        return decorator_solve
    else:
        return decorator_solve(fprime)


def convert_unit_to(expr, target_units, unit_system=kamodo_unit_system, raise_errors=True):
    """
    Same as sympy.convert_to but accepts equations and allows functions of units to pass

    Convert ``expr`` to the same expression with all of its units and quantities
    represented as factors of ``target_units``, whenever the dimension is compatible.

    ``target_units`` may be a single unit/quantity, or a collection of
    units/quantities.

    Examples
    ========

    >>> from sympy.physics.units import speed_of_light, meter, gram, second, day
    >>> from sympy.physics.units import mile, newton, kilogram, atomic_mass_constant
    >>> from sympy.physics.units import kilometer, centimeter
    >>> from sympy.physics.units import gravitational_constant, hbar
    >>> from sympy.physics.units import convert_to
    >>> convert_to(mile, kilometer)
    25146*kilometer/15625
    >>> convert_to(mile, kilometer).n()
    1.609344*kilometer
    >>> convert_to(speed_of_light, meter/second)
    299792458*meter/second
    >>> convert_to(day, second)
    86400*second
    >>> 3*newton
    3*newton
    >>> convert_to(3*newton, kilogram*meter/second**2)
    3*kilogram*meter/second**2
    >>> convert_to(atomic_mass_constant, gram)
    1.660539060e-24*gram

    Conversion to multiple units:

    >>> convert_to(speed_of_light, [meter, second])
    299792458*meter/second
    >>> convert_to(3*newton, [centimeter, gram, second])
    300000*centimeter*gram/second**2

    Conversion to Planck units:

    >>> from sympy.physics.units import gravitational_constant, hbar
    >>> convert_to(atomic_mass_constant, [gravitational_constant, speed_of_light, hbar]).n()
    7.62963085040767e-20*gravitational_constant**(-0.5)*hbar**0.5*speed_of_light**0.5

    """


    from sympy.physics.units import UnitSystem
    unit_system = UnitSystem.get_unit_system(unit_system)

    if not isinstance(target_units, (Iterable, Tuple)):
        target_units = [target_units]


    if hasattr(expr, 'rhs'):
        return Eq(convert_unit_to(expr.lhs, target_units, unit_system),
            convert_unit_to(expr.rhs, target_units, unit_system))
    # if type(type(expr)) is UndefinedFunction:
    if is_function(expr):
        # print('undefined input expr:{}'.format(expr))
        return nsimplify(expr, rational=True)

    if isinstance(expr, Add):
        return Add.fromiter(convert_unit_to(i, target_units, unit_system) for i in expr.args)

    expr = sympify(expr)

    if not isinstance(expr, Quantity) and expr.has(Quantity):
        try:
            expr = expr.replace(
                lambda x: isinstance(x, Quantity),
                lambda x: x.convert_to(target_units, unit_system))
        except OSError:
            raise OSError('problem converting {} to {}\n{}'.format(
                expr, target_units, unit_system))

    def get_total_scale_factor(expr):
        if isinstance(expr, Mul):
            return reduce(lambda x, y: x * y, [get_total_scale_factor(i) for i in expr.args])
        elif isinstance(expr, Pow):
            return get_total_scale_factor(expr.base) ** expr.exp
        elif isinstance(expr, Quantity):
            return unit_system.get_quantity_scale_factor(expr)
        return expr

    if expr == target_units:
        return expr

    depmat = _get_conversion_matrix_for_expr(expr, target_units, unit_system)
    if depmat is None:
        if raise_errors:
            raise NameError('cannot convert {} to {} {}'.format(expr, target_units, unit_system))
        return nsimplify(expr, rational=True)

    expr_scale_factor = get_total_scale_factor(expr)
    result = expr_scale_factor * Mul.fromiter(
        (1/get_total_scale_factor(u) * u) ** p for u, p in zip(target_units, depmat))
    return nsimplify(result, rational=True)


def get_expr_unit(expr, unit_registry, verbose=False):
    '''Get units from an expression'''
    if is_function(expr):
        for func in unit_registry:
            if type(expr) == type(func):
                # b(a) = b(x)
                if verbose:
                    print('get_expr_unit: found match {} for {}'.format(func, expr))
                # {f(x): f(cm), f(cm): kg}
                func_units = unit_registry[func]
                if func_units in unit_registry:
                    return unit_registry[func_units]
                return func_units
        if verbose:
            print('get_expr_unit: no match found for {}'.format(expr))

    if isinstance(expr, Mul):
        results = []
        for arg in expr.args:
            result = get_expr_unit(arg, unit_registry, verbose)
            if result is not None:
                results.append(result)
        if len(results) > 0:
            return Mul.fromiter(results)
        else:
            return None

    if len(unit_registry) > 0:
        expr_unit = expr.subs(
            unit_registry, simultaneous=False).subs(
                unit_registry, simultaneous=False)
    else:
        expr_unit = expr

    if len(expr_unit.free_symbols) > 0:
        return None

    if isinstance(expr_unit, Add):
        # use the first term
        arg_0 = expr_unit.args[0]
        convert_unit_to(expr_unit, arg_0, kamodo_unit_system)
        result = arg_0
    else:
        result = expr_unit

    # if len(result.free_symbols) > 0:
    #     return None

    return get_base_unit(result)


def get_arg_units(expr, unit_registry, arg_units=None):
    """retrieves units of expression arguments
    """
    if arg_units is None:
        arg_units = dict()

    if expr in unit_registry:
        # unit_registry: {f(cm,km): kg}
        if is_function(unit_registry[expr]):
            for arg_, arg_unit in zip(expr.args, unit_registry[expr].args):
                arg_units[arg_] = arg_unit
        return arg_units

    # 2*a(x)
    for arg in expr.args:
        arg_units = get_arg_units(arg, unit_registry, arg_units)
    return arg_units

def replace_args(expr, from_map, to_map):
    # func_symbol = type(expr)
    arg_map = dict()
    for arg in expr.free_symbols:
        if (arg in from_map) & (arg in to_map):
            from_unit = from_map[arg]
            to_unit = to_map[arg]
            try:
                assert get_dimensions(from_unit) == get_dimensions(to_unit)
                arg_map[arg] = convert_unit_to(
                    arg*to_unit,
                    from_unit,
                    kamodo_unit_system)/from_unit
            except:
                raise NameError('cannot convert from {} to {}'.format(from_unit, to_unit))
    return expr.subs(arg_map)

def unify_args(expr, unit_registry, to_symbol, verbose):
    """replaces arguments in an expression"""
    expr_units = get_arg_units(expr, unit_registry)
    to_units = get_arg_units(to_symbol, unit_registry)
    if verbose:
        print('unify_args: {} arg units {}'.format(expr, expr_units))
        print('unify_args: to arg units', to_units)
    expr = replace_args(expr, expr_units, to_units)
    if verbose:
        print('unify_args: converted expression:', expr)
    return expr

def unify(expr, unit_registry, to_symbol=None, verbose=False):
    """adds unit conversion factors to composed functions"""
    if verbose:
        print('unify: to_symbol:', to_symbol)
    if hasattr(expr, 'rhs'):
        return Eq(expr.lhs, unify(
            expr.rhs,
            unit_registry,
            to_symbol=expr.lhs,
            verbose=verbose))
    if isinstance(expr, Add):
        if verbose:
            print('unify: Adding expression: {} -> {}'.format(
                expr, get_expr_unit(to_symbol, unit_registry, verbose)))
        return Add.fromiter([unify(arg, unit_registry, to_symbol, verbose) for arg in expr.args])

    expr_unit = get_expr_unit(expr, unit_registry, verbose)

    if verbose:
        print('unify: {} unit {}'.format(expr, expr_unit))
        print('unify: {} symbols {}'.format(expr, expr.free_symbols))
        print('unify: {} symbols {}'.format(to_symbol, to_symbol.free_symbols))
    try:
        assert expr.free_symbols.issubset(to_symbol.free_symbols)
    except:
        raise NameError("{} arguments not in {}".format(
            expr.free_symbols, to_symbol.free_symbols))

    if is_function(expr):
        if verbose:
            print('unify: function expression: {}'.format(expr))
        if to_symbol is not None:
            if verbose:
                print('\nunify: to_symbol args: {}'.format(to_symbol.args))
                print('unify: to_symbol free symbols: {}'.format(to_symbol.free_symbols))
                print('unify: expr args: {}'.format(expr.args))
                print('unify: expr free symbols: {}'.format(expr.free_symbols))

            for k, v in unit_registry.items():
                if isinstance(expr, type(k)):
                    if len(k.free_symbols) > 0:
                        if verbose:
                            print('unify: found matching {} -> {}'.format(expr, k))
                        arg_units = get_arg_units(k, unit_registry)
                        if verbose:
                            print('unify: func units:', arg_units)
                            print('unify: {}->{}'.format(expr.args, k.args))
                        expr_units = {}
                        for arg, sym in zip(expr.args, k.args):
                            to_unit = arg_units.get(sym)
                            from_unit = get_expr_unit(arg, unit_registry)
                            if (from_unit is not None) and (to_unit is not None):
                                expr_units[arg] = convert_unit_to(
                                    arg*from_unit,
                                    to_unit,
                                    kamodo_unit_system)/to_unit
                        expr = expr.subs(expr_units)

                        if verbose:
                            print('unify: replaced args', expr)

    expr = unify_args(expr, unit_registry, to_symbol, verbose)

    if (to_symbol is not None) & (expr_unit is not None):
        to_unit = get_expr_unit(to_symbol, unit_registry, verbose)
        if verbose:
            print('unify: to_unit {}'.format(to_unit))
        expr_dimensions = get_dimensions(expr_unit)
        to_dimensions = get_dimensions(to_unit)
        if expr_dimensions.compare(to_dimensions) == 0:
            if verbose:
                print('unify: {} [{}] -> to_symbol: {}[{}]'.format(
                    expr, expr_unit, to_symbol, to_unit))
            expr = convert_unit_to(
                expr*expr_unit,
                to_unit,
                kamodo_unit_system)/to_unit
        else:
            if verbose:
                print('unify: registry:')
                for k, v in unit_registry.items():
                    print('unify:\t{} -> {}'.format(k, v))
                print('compare:{}'.format(expr_dimensions.compare(to_dimensions)))
                error_msg = 'cannot convert {} [{}] {} to {}[{}] {}'.format(
                expr, expr_unit, expr_dimensions,
                to_symbol, to_unit, to_dimensions)
                print(error_msg)
            raise NameError(error_msg)

    return expr

def get_abbrev(unit):
    """get the abbreviation for a mixed unit"""
    if hasattr(unit, 'abbrev'):
        return unit.abbrev
    if isinstance(unit, Mul):
        return Mul.fromiter([get_abbrev(arg) for arg in unit.args])
    if isinstance(unit, Pow):
        base, exp = unit.as_base_exp()
        return Pow(get_abbrev(base), exp)
    return unit


def base_dimensions(d):
    dependencies = dimsys_SI.get_dimensional_dependencies(d)
    return Mul.fromiter(Pow(Dimension(base), exp_) for base, exp_ in dependencies.items())

def get_dimensions(unit):
    """get the set of basis units"""
    if hasattr(unit, 'dimension'):
        base_dims = base_dimensions(unit.dimension)
        if len(base_dims.args) == 1:
            return base_dims.args[0]
        else:
            return Mul.fromiter([arg.args[0] for arg in base_dimensions(unit.dimension).args])
    if isinstance(unit, Mul):
        terms = [get_dimensions(arg) for arg in unit.args]
        return Mul.fromiter(terms)
    if isinstance(unit, Pow):
        base, exp = unit.as_base_exp()
        return Pow(get_dimensions(base), exp)
    return 1


def get_base_unit(expr):
    if hasattr(expr, 'dimension'):
        return expr
    if isinstance(expr, Mul):
        results = []
        for arg in expr.args:
            result = get_base_unit(arg)
            if result is not None:
                results.append(result)
        if len(results) > 0:
            return Mul.fromiter(results)
        else:
            return None
    if isinstance(expr, Pow):
        base, exp = expr.as_base_exp()
        return Pow(get_base_unit(base), exp)
    if isinstance(expr, Add):
        results = [get_dimensions(arg) for arg in expr.args]
        arg0_dimensions = results[0]
        for r in results:
            try:
                assert arg0_dimensions == r
            except:
                print(results)
                raise
        return get_base_unit(expr.args[0])
    return None

def is_function(expr):
    """returns True if expr is a function

    examples:
        f(x): True
        symbols('f', cls=UndefinedFunction): True
        x: False
    """
    if isinstance(expr, UndefinedFunction):
        return True
    return isinstance(type(expr), UndefinedFunction)


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)

class NumpyArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyArrayEncoder, self).default(obj)

def full_classname(o):
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__
    return module + '.' + o.__class__.__name__

def serialize(obj):
    if isinstance(obj, (np.ndarray, np.generic)):
        if isinstance(obj, np.ndarray):
            return {
                # '__ndarray__': base64.b64encode(obj.tobytes()),
                '__ndarray__': obj.tolist(),
                'dtype': obj.dtype.str,
                'shape': obj.shape,
            }
        elif isinstance(obj, (np.bool_, np.number)):
            return {
                # '__npgeneric__': base64.b64encode(obj.tobytes()),
                '__ndarray__': obj.tolist(),
                'dtype': obj.dtype.str,
            }
    if isinstance(obj, pd.DataFrame):
        return {
            '__pddataframe__': obj.values.tolist(),
            '__index__': serialize(obj.index),
            'dtype': full_classname(obj),
        }
    if isinstance(obj, pd.Index):
        if isinstance(obj, pd.DatetimeIndex):
            return {
                '__datetime__': [_ for _ in map(datetime.isoformat, obj)],
                'dtype': full_classname(obj),
            }
        else:
            return {
                '__index__': obj.tolist(),
                'dtype': full_classname(obj),
            }
    if isinstance(obj, pd.Series):
        return {
            '__pdseries__': obj.values.tolist(),
            '__index__': serialize(obj.index),
            'dtype': full_classname(obj),
        }
    if isinstance(obj, set):
        return {'__set__': list(obj)}
    if isinstance(obj, tuple):
        return {'__tuple__': list(obj)}
    if isinstance(obj, complex):
        return {'__complex__': obj.__repr__()}

    if isinstance(obj, types.GeneratorType):
        return {'__lambdagen__': [{
            'params':{k:serialize(v) for k, v in get_defaults(func).items()},
            'result':serialize(func())} for func in obj]}

    if isinstance(obj, int):
        return obj

    # Let the base class default method raise the TypeError
    raise TypeError('Unable to serialise object of type {}'.format(type(obj)))

def lambdagen(obj):
    """create a generator of lambda functions"""
    for func_ in obj['__lambdagen__']:
        signature = []
        for arg, arg_default in func_['params'].items():
            signature.append(forge.arg(
                arg,
                default=deserialize(arg_default)))
        @forge.sign(*signature)
        def func(*args, **kwargs):
            """API function"""
            return deserialize(func_['result'])

        yield kamodofy(func)

def deserialize(obj):
    # convert obj into numpy, pandas
    if isinstance(obj, dict):
        if '__ndarray__' in obj:
            # return np.frombuffer(
            #     base64.b64decode(obj['__ndarray__']),
            #     dtype=np.dtype(obj['dtype'])
            # ).reshape(obj['shape'])
            return np.array(obj['__ndarray__'])
        if '__npgeneric__' in obj:
            # return np.frombuffer(
            #     base64.b64decode(obj['__npgeneric__']),
            #     dtype=np.dtype(obj['dtype'])
            # )[0]
            return np.array(obj['__npgeneric__'])
        if '__pddataframe__' in obj:
            return pd.DataFrame(
                obj['__pddataframe__'],
                index=deserialize(obj['__index__']))
        if '__pdseries__' in obj:
            return pd.Series(
                obj['__pdseries__'],
                index=deserialize(obj['__index__']))
        if '__datetime__' in obj:
            return pd.to_datetime(obj['__datetime__'])
        if '__index__' in obj:
            return pd.Index(obj['__index__'])
        if '__set__' in obj:
            return set(obj['__set__'])
        if '__tuple__' in obj:
            return tuple(obj['__tuple__'])
        if '__complex__' in obj:
            return complex(obj['__complex__'])
        if '__lambdagen__' in obj:
            return lambdagen(obj)

    return obj

# over-write the load(s)/dump(s) functions
def load(*args, **kwargs):
    kwargs['object_hook'] = deserialize
    return json.load(*args, **kwargs)


def loads(*args, **kwargs):
    kwargs['object_hook'] = deserialize
    return json.loads(*args, **kwargs)


def dump(*args, **kwargs):
    kwargs['default'] = serialize
    return json.dump(*args, **kwargs)


def dumps(*args, **kwargs):
    kwargs['default'] = serialize
    return json.dumps(*args, **kwargs)

def get_undefined_funcs(expr):
    """retrieve an expression's undefined functions"""
    return expr.atoms(sympy.function.AppliedUndef)

def sign_defaults(symbol, expr, composition):
    '''gets defaults from an expression using composition'''
    defaults={}
    for f_ in get_undefined_funcs(expr):
        # includes {h(f(g)), f(g)}
        if str(f_) in composition:
            # ignores h(f(g))
            f_defaults = get_defaults(composition[str(f_)])
            # flatten defaults
            for arg, arg_default in f_defaults.items():
                defaults[arg] = arg_default

    arg_signatures = []
    # defaults have to go last, which may conflict with user's ordering
    for arg in symbol.args:
        str_arg = str(arg)
        if str_arg in defaults:
            arg_default=defaults[str(arg)]
            arg_signatures.append(forge.arg(str_arg, default=arg_default))
        else:
            arg_signatures.append(forge.arg(str_arg))

    # will raise an error if defaults are not last
    signature = forge.sign(*arg_signatures)
    return signature

class LambdaGenerator(object):
    def __init__(self, lambda_generator):
        """supports simple expressions for manipulating lambda generators"""
        self._lambda_generator = lambda_generator

    def __add__(self, other):
        if isinstance(other, LambdaGenerator):
            for func, gunc in zip(self._lambda_generator, other._lambda_generator):
                yield lambda: func()+gunc()
        else:
            for func in self._lambda_generator:
                yield lambda: func()+other
    def __sub__(self, other):
        if isinstance(other, LambdaGenerator):
            for func, gunc in zip(self._lambda_generator, other._lambda_generator):
                yield lambda: func()-gunc()
        else:
            for func in self._lambda_generator:
                yield lambda: func()-other
    def __mul__(self, other):
        if isinstance(other, LambdaGenerator):
            for func, gunc in zip(self._lambda_generator, other._lambda_generator):
                # what do we do with defaults?
                yield lambda: func()*gunc()
        else:
            for func in self._lambda_generator:
                yield lambda: func()*other
    def __truediv__(self, other):
        if isinstance(other, LambdaGenerator):
            for func, gunc in zip(self._lambda_generator, other._lambda_generator):
                # what do we do with defaults?
                yield lambda: func().__truediv__(gunc())
        else:
            for func in self._lambda_generator:
                yield lambda: func().__truediv__(other)
    def __floordiv__(self, other):
        if isinstance(other, LambdaGenerator):
            for func, gunc in zip(self._lambda_generator, other._lambda_generator):
                # what do we do with defaults?
                yield lambda: func().__floordiv__(gunc())
        else:
            for func in self._lambda_generator:
                yield lambda: func().__floordiv__(other)
    def __pow__(self, other):
        if isinstance(other, LambdaGenerator):
            for func, gunc in zip(self._lambda_generator, other._lambda_generator):
                yield lambda: func().__pow__(gunc())
        else:
            for func in self._lambda_generator:
                yield lambda: func().__pow__(other)


def get_bbox(fig):
    t0 = fig['data'][0]
    xmin = np.nanmin(t0['x'])
    xmax = np.nanmax(t0['x'])
    ymin = np.nanmin(t0['y'])
    ymax = np.nanmax(t0['y'])
    if 'z' in t0:
        zmin = np.nanmin(t0['z'])
        zmax = np.nanmax(t0['z'])
        for t in fig['data']:
            xmin = min(xmin, np.nanmin(t['x']))
            xmax = max(xmax, np.nanmax(t['x']))
            ymin = min(ymin, np.nanmin(t['y']))
            ymax = max(ymax, np.nanmax(t['y']))
            zmin = min(zmin, np.nanmin(t['z']))
            zmax = max(zmax, np.nanmax(t['z']))
        return xmin, xmax, ymin, ymax, zmin, zmax
    else:
        for t in fig['data']:
            xmin = min(xmin, np.nanmin(t['x']))
            xmax = max(xmax, np.nanmax(t['x']))
            ymin = min(ymin, np.nanmin(t['y']))
            ymax = max(ymax, np.nanmax(t['y']))
        return xmin, xmax, ymin, ymax


def set_aspect(fig):
    """sets aspect ratio of the scene based on bounding box of traces"""
    fig.layout.scene.aspectmode='manual'
    xmin, xmax, ymin, ymax, zmin, zmax = get_bbox(fig)
    fig.layout.scene.aspectratio=dict(x=xmax-xmin, y=ymax-ymin,z=zmax-zmin)
    return fig

def curry(func):
    """currying function

    borrowed from https://www.python-course.eu/currying_in_python.php
    """
    # to keep the name of the curried function:

    curry.__curried_func_name__ = func.__name__
    f_args, f_kwargs = [], {}

    def f(*args, **kwargs):
        """a curried function"""

        nonlocal f_args, f_kwargs
        if args or kwargs:
            f_args += args
            f_kwargs.update(kwargs)
            return f
        else:
            result = func(*f_args, **f_kwargs)
            f_args, f_kwargs = [], {}
            return result
    return f


def construct_signature(*args, **kwargs):
    """construct a signature
    usage:
    
        @forge.sign(*get_signature('x','y',z=3))
        def f(*args, **kargs):
            pass
    """
    signature = []
    for arg in args:
        signature.append(forge.arg(arg))
    for k, v in kwargs.items():
        signature.append(forge.arg(k, default=v))
    return signature

def partial(_func=None, **partial_kwargs):
    """A partial function decorator

    Reduces function signature to reflect partially assigned kwargs
    """
    verbose = partial_kwargs.pop('verbose', False)

    def decorator_partial(f):
        orig_args = get_args(f)
        orig_defaults = get_defaults(f)
        orig_defaults.update(partial_kwargs)
        if verbose:
            print('partial kwargs', partial_kwargs)
            print('original args:', orig_args)
            print('new defaults', orig_defaults)

        # collect only the arguments not assigned by partial
        sig_defaults = {}
        sig_args = []
        for arg in orig_args:
            if arg in partial_kwargs:
                continue
            if arg in orig_defaults:
                sig_defaults[arg] = orig_defaults[arg]
            else:
                sig_args.append(arg)

        if verbose:
            print('updated signature:', sig_args, sig_defaults)

        @forge.sign(*construct_signature(*sig_args, **sig_defaults))
        def wrapped(*args, **kwargs):
            """simple wrapper"""
            kwargs.update(partial_kwargs)
            if verbose:
                print('kwargs to pass:', kwargs)
            return f(*args, **kwargs)

        if verbose:
            print('wrapped docs', wrapped.__doc__)
        wrapped = decorate(wrapped, decorator_wrapper)

        wrapped.__name__ = f.__name__
        orig_docs = f.__doc__
        for k, v in sig_defaults.items():
            sig_args.append('{}={}'.format(k, v))
        doc_args = ', '.join(sig_args)
        rhs_args = []
        for arg in orig_args:
            if arg in partial_kwargs:
                rhs_args.append('{}={}'.format(arg, partial_kwargs[arg]))
            else:
                rhs_args.append(arg)

        doc_orig_args = ', '.join(rhs_args)
        wrapped.__doc__ = """Calling {fname}({orig_args}) for fixed {partial_keys}:
        {fname}({doc_args}) = {fname}({orig_args_fixed})\n""".format(
            orig_args_fixed=doc_orig_args,
            doc_args=doc_args,
            fname=f.__name__,
            orig_args = ', '.join(orig_args),
            partial_keys=', '.join(partial_kwargs.keys()),
        )
        
        if orig_docs is not None:
            wrapped.__doc__ += "\n" + f.__doc__

        return wrapped

    if _func is None:
        return decorator_partial
    else:
        return decorator_partial(_func)



# -*- coding: utf-8 -*-
"""
Copyright © 2017 United States Government as represented by the Administrator, National Aeronautics and Space Administration.  
No Copyright is claimed in the United States under Title 17, U.S. Code.  All Other Rights Reserved.
"""

import numpy as np
from sympy import Integral, Symbol, symbols, Function

# try:
#     from sympy.parsing.sympy_parser import parse_expr
# except ImportError:  # may occur for certain versions of sympy
#     from sympy import sympify as parse_expr

from collections import OrderedDict
from collections import UserDict
import collections

from sympy import lambdify
from sympy.parsing.latex import parse_latex
from sympy import latex
from sympy.core.function import UndefinedFunction
from inspect import getfullargspec
from sympy import Eq
import pandas as pd
# from IPython.display import Latex


from sympy.physics import units as sympy_units
from sympy.physics.units import Quantity
from sympy.physics.units import Dimension
from sympy import Expr

import functools
from .util import kamodofy
from .util import sort_symbols
from .util import simulate
from .util import unit_subs
from .util import get_defaults, valid_args, eval_func
# from .util import to_arrays, cast_0_dim
from .util import beautify_latex, arg_to_latex
from .util import concat_solution
from .util import convert_unit_to
from .util import unify, get_abbrev, get_expr_unit
from .util import is_function, get_arg_units


import plotly.graph_objs as go
from plotly import figure_factory as ff

from .plotting import plot_dict, get_arg_shapes, symbolic_shape
from .plotting import get_ranges
from .util import existing_plot_types
from .util import get_dimensions
from .util import reserved_names

from sympy import Wild
from types import GeneratorType
import inspect

import re

import urllib.request, json
import requests
from .util import serialize, deserialize
from .util import sign_defaults
import forge
from sympy.abc import _clash
import sympy

_clash['rad'] = Symbol('rad')
_clash['deg'] = Symbol('deg')



def parse_expr(*args, **kwargs):
    try:
        return sympy.sympify(*args, **kwargs)
    except:
        raise NameError('cannot parse {}, {}'.format(args, kwargs))


def get_unit_quantities():
    """returns a map of sympy units {symbol: quantity} """
    subs = {}
    for k, v in list(sympy_units.__dict__.items()):
        if isinstance(v, Expr) and v.has(sympy_units.Unit):
            subs[Symbol(k)] = v  # Map the `Symbol` for a unit to the quantity
    return subs


def clean_unit(unit_str):
    """remove brackets and all white spaces in and around unit string"""
    return unit_str.strip().strip('[]').strip()


def get_unit(unit_str, unit_subs=unit_subs):
    """get the unit quantity corresponding to this string

    unit_subs should contain a dictonary {symbol: quantity} of custom units not available in sympy"""

    unit_str = clean_unit(unit_str)
    units = get_unit_quantities()

    if unit_subs is not None:
        units.update(unit_subs)

    if len(unit_str) == 0:
        return Dimension(1)

    unit_expr = parse_expr(unit_str.replace('^', '**'), locals=_clash)
    try:
        unit = unit_expr.subs(units)
    except:
        raise NameError('something wrong with unit str [{}], type {}, {}'.format(
            unit_str, type(unit_str), unit_expr))
    try:
        assert len(unit.free_symbols) == 0
    except:
        raise NameError("Unsupported unit: {} {} {}".format(
            unit_str, type(unit_str), unit.free_symbols))
    return unit




def args_from_dict(expr, local_dict, verbose):
    """retrieve the symbols of an expression

    if a  {str: symbol} mapping is provided, use its symbols instead
    """
    args = []
    # if local_dict is not None:
    if verbose:
        print('args_from_dict: available names {}'.format(local_dict.keys()))
    for a in expr.args:
        if verbose:
            print('args_from_dict: {}'.format(a))
        if str(a) in local_dict:
            args.append(local_dict[str(a)])
        else:
            args.append(a)
    return tuple(args)
    # else:
    #     return expr.args

def get_str_units(bracketed_str):
    """finds all bracketed units in a string

    supports functions like
        bracketed_str = 'f(x[m],y[cm**3],z[km^.5],q[(cm*g)^-3])[km/s]'
        get_str_units(bracketed_str)
        >>> ['m', 'cm**3', 'km^.5', '(cm*g)^-3', 'km/s']
    """
    return re.findall(r"\[([A-Za-z0-9_/*^\\(\\)-\.]+)\]", bracketed_str)

def str_has_arguments(func_str): 
    bracket = func_str.find('[')
    paren = func_str.find('(')
    if bracket > 0:
        if paren > 0:
            # f(x[(cm)^2]) has arguments since paren comes first
            # f[(cm)^2] has no arguments since bracket comes first
            return bracket > paren
        return False # no paren hence no arguments
    # no bracket
    if paren > 0:
        # has argument
        return True
    return False


def extract_units(func_str):
    """separate the units from the left-hand-side of an expression assuming bracket notation

    We return symbol and dictionary mapping symbols to units

    needs to handle the following cases:

        args and output have units
            extract_units('f(x[cm],y[km])[kg]')
            ('f(x,y)', {'x': 'cm', 'y': 'km', 'f(x,y)': 'kg'})

        args have parenthesis in units and output has no units
            extract_units('f(x[(cm )^2])')
            ('f(x)', {'x': '(cm)^2', 'f(x)': ''})

        output has parenthesis in units
            extract_units('f[(cm)^2]')
            ('f', {'f': '(cm)^2'})

        no args named and no units named
            ('f', {'f': ''})

    """
    # remove any spaces
    func_str = func_str.replace(' ', '')
    all_units = get_str_units(func_str)

    unit_dict = {}
    if str_has_arguments(func_str):
        # f(x[cm],y[km])[kg] -> x[cm],y[km]
        lhs_args = re.findall(r"\((.+)\)", func_str)[0]

        # 'x[cm],y[km]' -> ['cm', 'km']
        arg_units = get_str_units(lhs_args)
        args = lhs_args.split(',')

        for arg, unit in zip(args, arg_units):
            unit_dict[arg.replace('[{}]'.format(unit), '')] = unit

        if len(all_units) == len(arg_units):
            output_units = ''
        else:
            output_units = all_units[-1]
    else:
        if len(all_units) > 0:
            output_units = all_units[0]
        else:
            output_units = ''


    # f(x[cm],y[km])[kg]
    lhs = func_str
    for unit in all_units:
        lhs = lhs.replace('[{}]'.format(unit), '')

    unit_dict[lhs] = output_units
    return lhs, unit_dict


def expr_to_symbol(expr, args):
    """f,args -> f(args) of class function, f(x) -> f(x)"""
    if type(expr) == Symbol:
        return parse_expr(str(expr) + str(args))
        # return symbols(str(expr), cls = Function)
    else:
        return expr


def parse_lhs(lhs, local_dict, verbose):
    """parse the left-hand-side

    returns: symbol, arguments, units, parsed_expr
    """
    lhs, unit_dict = extract_units(lhs)
    if lhs in reserved_names:
        raise NameError('{} is a reserved name'.format(lhs))
    parsed = parse_expr(lhs)
    args = args_from_dict(parsed, local_dict, verbose)
    symbol = expr_to_symbol(parsed, args)
    return symbol, args, unit_dict, parsed


def parse_rhs(rhs, is_latex, local_dict):
    """parse the right-hand-side of an equation

    subsititute variables from local_dict where available
    """
    if is_latex:
        expr = parse_latex(rhs).subs(local_dict)
    else:
        try:
            expr = parse_expr(rhs).subs(local_dict)
        except SyntaxError:
            print('cannot parse {} with {}'.format(rhs, local_dict))
            raise
    return expr


def get_function_args(func, hidden_args=[]):
    """converts function arguments to list of symbols"""
    return symbols([a for a in getfullargspec(func).args if a not in hidden_args])




# class Kamodo(collections.OrderedDict):
class Kamodo(UserDict):
    '''Kamodo base class demonstrating common API for space weather models


    This API provides access to space weather fields and their properties through:

        interpolation of variables at user-defined points
        unit conversions
        coordinate transformations specific to space weather domains

    Required methods that have not been implemented in child classes
    will raise a NotImplementedError
    '''

    def __init__(self, *funcs, **kwargs):
        """Base initialization method

        Args:
            param1 (str, optional): Filename of datafile to interpolate from

        """

        super(Kamodo, self).__init__()
        self.symbol_registry = OrderedDict()
        self.unit_registry = OrderedDict()

        symbol_dict = kwargs.pop('symbol_dict', None)

        self.verbose = kwargs.pop('verbose', False)
        self.signatures = OrderedDict()

        for func in funcs:
            if self.verbose:
                print('registering {}'.format(func))
            if type(func) is str:
                components = func.split('=')
                if len(components) == 2:
                    # function(arg)[unit] = expr
                    lhs, rhs = components
                    self[lhs.strip('$')] = rhs
                else:
                    raise NotImplementedError(
                        'cannot register functions of the form {}'.format(func))

        for sym_name, expr in list(kwargs.items()):
            if self.verbose:
                print('registering {} with {}'.format(sym_name, expr))
            self[sym_name] = expr


    def register_symbol(self, symbol):
        self.symbol_registry[str(type(symbol))] = symbol


    # def remove_symbol(self, sym_name):
    #     """remove the sym_name (str) from the object"""

    #     if self.verbose:
    #         print('remove_symbol: removing {} from symbol_registry'.format(sym_name))
    #     try:
    #         symbol = self.symbol_registry.pop(sym_name)
    #     except KeyError:
    #         raise KeyError('{} was not in symbol_registry {}'.format(
    #             sym_name, self.symbol_registry.keys()))
    #     if self.verbose:
    #         print('remove_symbol: removing {} from signatures'.format(symbol))
    #     self.signatures.pop(str(type(symbol)))
    #     if self.verbose:
    #         print('remove_symbol: removing {} from unit_registry'.format(symbol))
    #     remove = []
    #     for sym in self.unit_registry:
    #         if is_function(sym): # {rho(x): rho(cm), rho(cm): kg}
    #             if type(sym) == type(symbol):
    #                 remove.append(sym)
    #     for sym in remove:
    #         self.unit_registry.pop(sym)

    #     if self.verbose:
    #         print('removing {} from {}'.format(symbol, self.keys()))

    #     if symbol in self.keys():
    #         try:
    #             self.pop(symbol)
    #         except KeyError:
    #             if self.verbose:
    #                 print('something wrong with {} in {}'.format(symbol, self.keys()))
    #             raise KeyError('{} not in {}'.format(symbol, self.keys()))

    #     if type(symbol) in self.keys():
    #         try:
    #             self.pop(type(symbol))
    #         except KeyError:
    #             raise KeyError('{} not in {}'.format(type(symbol), self.keys()))


    def parse_key(self, sym_name):
        """parses the symbol name

        sym_name must be a string

        returns: symbol, args, unit_dict, lhs_expr
        """
        args = tuple()
        try:
            symbol, args, unit_dict, lhs_expr = parse_lhs(
                sym_name,
                self.symbol_registry,
                self.verbose)
        except:
            raise NotImplementedError('could not parse key {}'.format(sym_name))

        # try:
        if len(args) > 0:
            symbol_str = str(symbol).replace(' ', '')
        else:
            symbol_str = str(type(symbol))
        units = get_abbrev(unit_dict[symbol_str])
        # except KeyError:
        #     raise KeyError('{} not in {}'.format(symbol_str, unit_dict))

        return symbol, args, units, lhs_expr

    def parse_value(self, rhs_expr, local_dict):
        """returns an expression from string"""
        if type(rhs_expr) is str:
            is_latex = '$' in rhs_expr
            rhs_expr = rhs_expr.strip('$').strip()
            rhs_expr = parse_rhs(rhs_expr, is_latex, local_dict)
        return rhs_expr

    def check_or_replace_symbol(self, symbol, free_symbols, rhs_expr=None):
        """Rules of replacement:

        """
        if self.verbose:
            print('symbol arguments: >>> {} {} <<<'.format(symbol.args, type(symbol.args)))
            print('free_symbols: >>> {} {} <<<'.format(free_symbols, type(free_symbols)))
        # try:
        lhs_args = set(symbol.args)
        # except TypeError:
        #     lhs_args = free_symbols
        #     symbol = Function(str(symbol))(*lhs_args)

        if lhs_args != set(free_symbols):
            free_symbols_ = tuple(free_symbols)
            if len(free_symbols) == 1:
                # try:
                symbol = parse_expr(str(type(symbol)) + str(free_symbols_))
                # except:
                #     raise NotImplementedError('cannot parse', str(type(symbol)) + str(free_symbols_))
            else:
                if len(lhs_args) > 0:
                    raise NameError("Mismatched symbols {} and {}".format(lhs_args, free_symbols))
                else:
                    if self.verbose:
                        print('type of rhs symbols:', type(free_symbols))
                    # if isinstance(free_symbols, set):
                    #     free_symbols_ = sort_symbols(free_symbols)
                    if self.verbose:
                        print('replacing {} with {}'.format(
                            symbol, str(type(symbol)) + str(free_symbols_)))
                    # try:
                    symbol = parse_expr(str(type(symbol)) + str(free_symbols_))
                    # except:
                    #     raise NotImplementedError('cannot parse', str(type(symbol)) + str(free_symbols_))

        return symbol

    def validate_function(self, lhs_expr, rhs_expr):
        assert lhs_expr.free_symbols == rhs_expr.free_symbols


    def vectorize_function(self, symbol, rhs_expr, composition):
        try:
            func = lambdify(symbol.args, rhs_expr, modules=['numexpr'])
            if self.verbose:
                print('lambda {} = {} labmdified with numexpr'.format(symbol.args, rhs_expr))
        except: # numexpr not installed
            func = lambdify(symbol.args, rhs_expr, modules=['numpy', composition])
            if self.verbose:
                print('lambda {} = {} lambdified with numpy and composition:'.format(
                    symbol.args, rhs_expr))
                for k, v in composition.items():
                    print('\t', k, v)
        signature = sign_defaults(symbol, rhs_expr, composition)
        return signature(func)


    def update_unit_registry(self, func, arg_units):
        """Inserts unit functions into registry"""
        lhs, unit_dict = extract_units(func)
        # if arg_units is None:
        #     arg_units = {}
        for key, value in unit_dict.items():
            if key != lhs:
                arg_units[parse_expr(key)] = get_unit(value)

        lhs_expr = parse_expr(lhs)
        func_units = lhs_expr.subs(arg_units)
        self.unit_registry[lhs_expr] = func_units
        output_units = unit_dict[lhs]
        self.unit_registry[func_units] = get_unit(output_units)
        return lhs


    def register_signature(self, symbol, units, lhs_expr, rhs_expr):
        # if isinstance(units, str):
        unit_str = units
        if self.verbose:
            print('unit str {}'.format(unit_str))

        self.signatures[str(type(symbol))] = dict(
            symbol=symbol,
            units=unit_str,
            lhs=lhs_expr,
            rhs=rhs_expr,
            # update = getattr(self[lhs_expr],'update', None),
        )

    def register_function(self, func, lhs_symbol, lhs_expr, lhs_units):
        hidden_args = []
        if hasattr(func, 'meta'):
            hidden_args = func.meta.get('hidden_args', [])

        if type(func) is np.vectorize:
            rhs_args = get_function_args(func.pyfunc, hidden_args)
        else:
            rhs_args = get_function_args(func, hidden_args)
        if str(rhs_args[0]) == 'self':  # in case function is a class method
            rhs_args.pop(0)

        lhs_symbol_old = lhs_symbol
        lhs_symbol = self.check_or_replace_symbol(lhs_symbol, rhs_args)
        units = lhs_units
        if hasattr(func, 'meta'):
            if self.verbose:
                print('function has metadata', func.meta.keys())
            if len(lhs_units) > 0:
                if lhs_units != func.meta['units']:
                    raise NameError('Units mismatch:{} != {}'.format(lhs_units, func.meta['units']))
            else:
                units = func.meta['units']
            rhs = func.meta.get('equation', func)
            if self.verbose:
                print('rhs from meta: {}'.format(rhs))
        else:
            rhs = func
            if self.verbose:
                print('rhs from input func {}'.format(rhs))
            try:
                setattr(func, 'meta', dict(units=lhs_units, arg_units=None))
            except:  # will not work on bound methods
                pass

        arg_units = func.meta.get('arg_units', None)
        if arg_units is not None:
            unit_dict = {arg: get_unit(arg_unit) for arg, arg_unit in arg_units.items()}
            unit_expr = lhs_symbol.subs(unit_dict)
            self.unit_registry[lhs_symbol] = unit_expr
            self.unit_registry[unit_expr] = get_unit(units)
        else:
            self.unit_registry[lhs_symbol] = get_unit(units)

        self.register_signature(lhs_symbol, units, lhs_expr, rhs)
        try:
            func._repr_latex_ = lambda: self.func_latex(str(type(lhs_symbol)), mode='inline')
        except AttributeError:
            # happens with bound methods
            pass
        super(Kamodo, self).__setitem__(lhs_symbol, func)  # assign key 'f(x)'
        super(Kamodo, self).__setitem__(type(lhs_symbol), self[lhs_symbol])  # assign key 'f'
        self.register_symbol(lhs_symbol)


    def __setitem__(self, sym_name, input_expr):
        """Assigns a function or expression to a new symbol,
        performs unit conversion where appropriate

        """
        if not isinstance(sym_name, str):
            sym_name = str(sym_name)

        symbol, args, lhs_units, lhs_expr = self.parse_key(sym_name)

        if hasattr(input_expr, '__call__'):
            self.register_function(input_expr, symbol, lhs_expr, lhs_units)

        else:
            if self.verbose:
                print(
                    "\n\nPARSING WITH UNIFY",
                    lhs_expr,
                    symbol,
                    lhs_units,
                    len(lhs_units),
                    type(lhs_units))
                print('symbol registry:', self.symbol_registry)

            rhs_expr = self.parse_value(input_expr, self.symbol_registry)

            if self.verbose:
                print('parsed rhs_expr', rhs_expr)

            if not isinstance(symbol, Symbol):
                if isinstance(lhs_expr, Symbol):
                    symbol = Function(lhs_expr)(*tuple(rhs_expr.free_symbols))
                else: #lhs is already a function
                    symbol = lhs_expr
                lhs_str = str(symbol)
                sym_name = sym_name.replace(str(lhs_expr), lhs_str)
            if self.verbose:
                print('unit registry contents:')
                for k, v in self.unit_registry.items():
                    print('\t', k, type(k), v)
            if '[' in sym_name:
                if self.verbose:
                    print('updating unit registry with {} -> {}'.format(sym_name, rhs_expr))
                rhs = rhs_expr
                arg_units = get_arg_units(rhs_expr, self.unit_registry)
                if self.verbose:
                    print(arg_units)
                sym_name = self.update_unit_registry(sym_name, arg_units)
                if self.verbose:
                    print('unit registry update returned', sym_name, self.unit_registry.get(symbol))
            else:

                if self.verbose:
                    print(sym_name,
                          symbol,
                          'had no units. Getting units from {}'.format(rhs_expr))

                expr_unit = get_expr_unit(rhs_expr, self.unit_registry, self.verbose)
                arg_units = get_arg_units(rhs_expr, self.unit_registry)


                if self.verbose:
                    print('registering {} with {} {}'.format(symbol, expr_unit, arg_units))

                if (symbol not in self.unit_registry) and (expr_unit is not None):
                    self.unit_registry[symbol] = symbol.subs(arg_units)
                    self.unit_registry[symbol.subs(arg_units)] = expr_unit


                if expr_unit is not None:
                    expr_dimensions = Dimension(get_dimensions(expr_unit))
                    if expr_dimensions != Dimension(1):
                        lhs_units = str(get_abbrev(get_expr_unit(
                            expr_unit,
                            self.unit_registry,
                            self.verbose)))
                    else:
                        lhs_units = ''

                if self.verbose:
                    print('registered lhs_units', lhs_units)

                rhs = rhs_expr
                sym_name = str(sym_name)



            if len(lhs_units) > 0:
                if self.verbose:
                    print('about to unify lhs_units {} {} with {}'.format(
                        lhs_units, type(lhs_units), rhs))

                expr = unify(
                    Eq(parse_expr(sym_name), rhs),
                    self.unit_registry,
                    # to_symbol = symbol,
                    verbose=self.verbose)
                rhs_expr = expr.rhs

            if self.verbose:
                print('symbol after unify', symbol, type(symbol), rhs_expr)
                print('unit registry to resolve units:')
                for k,v in self.unit_registry.items():
                    print('\t{}:{}'.format(k,v))

            units = get_expr_unit(symbol, self.unit_registry)
            if Dimension(get_dimensions(units)) != Dimension(1):
                units = get_abbrev(units)
                if units is not None:
                    units = str(units)
                else:
                    units = ''
            else:
                units = ''

            if self.verbose:
                print('units after resolve', symbol, units)
                for k, v in self.unit_registry.items():
                    print('\t{}: {}'.format(k, v))

            rhs_args = rhs_expr.free_symbols

            symbol = self.check_or_replace_symbol(symbol, rhs_args, rhs_expr)
            self.validate_function(symbol, rhs_expr)

            composition = {str(k_): self[k_] for k_ in self}
            arg_units = {}
            if symbol in self.unit_registry:
                unit_args = self.unit_registry[symbol]
                if unit_args is not None:
                    if len(unit_args.args) == len(symbol.args):
                        for arg, unit in zip(symbol.args, unit_args.args):
                            arg_units[str(arg)] = str(get_abbrev(unit))
            func = self.vectorize_function(symbol, rhs_expr, composition)
            meta = dict(units=units, arg_units=arg_units)
            func.meta = meta
            func.data = None
            self.register_signature(symbol, units, lhs_expr, rhs_expr)
            func._repr_latex_ = lambda: self.func_latex(str(type(symbol)), mode='inline')
            super(Kamodo, self).__setitem__(symbol, func)
            super(Kamodo, self).__setitem__(type(symbol), self[symbol])
            self.register_symbol(symbol)


    def __getitem__(self, key):
        try:
            return super(Kamodo, self).__getitem__(key)
        except KeyError:
            try:
                symbol = self.symbol_registry[str(key)]
            except KeyError:
                symbol = self.symbol_registry[str(type(key))]
            if symbol in self:
                return super().__getitem__(symbol)

    def __contains__(self, item):
        func_str = str(item).replace(' ', '')
        for key in self.keys():
            if func_str == str(key).replace(' ', ''):
                return True
        return False

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __delattr__(self, name):

        if name in self:
            symbol = self.symbol_registry[name]
            if self.verbose:
                print('__delattr__: removing name {}, symbol {}'.format(name, symbol))
                print(self.keys())
            # del self[symbol]
            super().pop(symbol)
            super().pop(type(symbol))

        else:
            raise AttributeError("No such field: " + name)

    def __delitem__(self, key):
        if self.verbose:
            print('__delitem__: got {} type: {}'.format(key, type(key)))
        if isinstance(key, str):
            if self.verbose:
                print('__delitem__: removing {} from symbol_registry'.format(key))
            symbol = self.symbol_registry.pop(key, None)
            if symbol is None:
                symbol, args, unit_dict, parsed = parse_lhs(
                    key, self.symbol_registry, self.verbose)
        else:
            symbol = key
            if self.verbose:
                print('__delitem__: received non str key {}'.format(symbol))
        if self.verbose:
            print('__delitem__: removing {} {}'.format(symbol, type(symbol)))


        remove_keys = []
        for k in self.data:
            if str(k) == str(symbol):
                remove_keys.append(k)
            if str(k) == str(type(symbol)):
                remove_keys.append(k)
        if self.verbose:
            print('__delitem__: removing keys:', remove_keys)

        for key_ in remove_keys:
            self.data.pop(key_)
            self.signatures.pop(str(key_), None)
            self.signatures.pop(str(type(key_)), None)
            self.unit_registry.pop(key_, None)

    # def get_units_map(self):
    #     """Maps from string units to symbolic units"""
    #     d = dict()
    #     star = Wild('star')
    #     for k, v in list(self.signatures.items()):
    #         unit = get_unit(v['units'])
    #         d[k] = unit
    #         d[v['symbol']] = unit

    #     return d

    def func_latex(self, key, mode='equation'):
        """get a latex string for a given function key"""
        repr_latex = ""
        lhs = self.signatures[key]['symbol']
        rhs = self.signatures[key]['rhs']
        units = self.signatures[key]['units']
        arg_units = get_arg_units(lhs, self.unit_registry)

        if len(units) > 0:
            units = '{}'.format(get_abbrev(units))
        else:
            units = ''

        if len(arg_units) > 0:
            arg_strs = []
            for arg, arg_unit in arg_units.items():
                arg_strs.append("{}[{}]".format(
                    latex(arg),
                    latex(get_abbrev(arg_unit),
                        fold_frac_powers=True,
                        fold_short_frac=True,
                        root_notation=False,
                        )))
            lhs_str = "{}({})".format(latex(type(lhs)), ",".join(arg_strs))
        else:
            lhs_str = latex(lhs)
            # lhs_str = "{}({})".format(
            #     latex(type(lhs)),
            #     ','.join([latex(s) for s in lhs.args]))
        if len(units) > 0:
            lhs_str += "[{}]".format(latex(parse_expr(units.replace('^', '**'), locals=_clash),
                    fold_frac_powers=True,
                    fold_short_frac=True,
                    root_notation=False,
                ))


        latex_eq = ''
        latex_eq_rhs = ''

        if isinstance(rhs, str):
            latex_eq_rhs = rhs       
        elif hasattr(rhs, '__call__') | (rhs is None):
            lambda_ = symbols('lambda', cls=UndefinedFunction)
            # latex_eq = latex(Eq(lhs, lambda_(*lhs.args)), mode=mode)
            latex_eq_rhs = latex(lambda_(*lhs.args)) # no $$
        else:
            # latex_eq = latex(Eq(lhs, rhs), mode=mode)
            latex_eq_rhs = latex(rhs) # no $$ delimiter

        if len(str(units)) > 0:
            latex_eq = latex_eq.replace('=', '[{}] ='.format(units))

        if mode == 'equation':
            repr_latex += r"\begin{equation}"
            repr_latex += "{} = {}".format(lhs_str, latex_eq_rhs)
            repr_latex += r"\end{equation}"
        else:
            repr_latex += r"$"
            repr_latex += "{} = {}".format(lhs_str, latex_eq_rhs)
            repr_latex += r"$"

        return repr_latex



    def to_latex(self, keys=None, mode='equation'):
        """Generate list of LaTeX-formated formulas
        Upon registeration, each function should have a _repr_latex_ method.
        """
        if keys is None:
            keys = list(self.signatures.keys())

        funcs_latex = []
        for k in keys:
            # funcs_latex.append(self.func_latex(k, mode))
            func_latex = self[k]._repr_latex_()
            if mode == 'equation':
                func_latex = func_latex.strip('$')
                func_latex = r"\begin{equation}" + func_latex + r"\end{equation}"
            funcs_latex.append(func_latex)

        repr_latex = " ".join(funcs_latex)

        return beautify_latex(repr_latex).encode('utf-8').decode()

    def _repr_latex_(self):
        """Provide notebook rendering of formulas"""
        return self.to_latex()

    def detail(self):
        """Constructs a pandas dataframe from signatures"""
        return pd.DataFrame(self.signatures).T


    def simulate(self, **kwargs):
        state_funcs = []
        for name, func_key in list(self.symbol_registry.items()):
            func = self[func_key]
            update_var = getattr(func, 'update', None)
            if update_var is not None:
                state_funcs.append((func.update, func))

        return simulate(OrderedDict(state_funcs), **kwargs)

    def evaluate(self, variable, *args, **kwargs):
        """evaluates the variable

        if the variable is not present, try to parse it as a semicolon-delimited list
        """
        if not hasattr(self, variable):
            var_dict = {}
            for variable_ in variable.split(';'):
                if len(variable_.split('=')) == 2:
                    variable_name, variable_expr = variable_.strip("'").split('=')
                    var_dict[variable_name] = variable_expr
                else:
                    raise SyntaxError('cannot parse {}'.format(variable_))
            knew = from_kamodo(self, **var_dict)
            # evaluate the last variable
            result = knew.evaluate(variable=variable_name, *args, **kwargs)
            return result


        if isinstance(self[variable], np.vectorize):
            params = get_defaults(self[variable].pyfunc)
        else:
            params = get_defaults(self[variable])

        valid_arguments = valid_args(self[variable], kwargs)
        if len(params) > 0:
            if self.verbose:
                print('default parameters:', list(params.keys()))
            params.update(valid_arguments)
        else:
            if self.verbose:
                print('no default parameters')
            params = valid_arguments
            if self.verbose:
                print('user-supplied args:', list(params.keys()))
        result = eval_func(self[variable], params)
        if isinstance(result, GeneratorType):
            if (len(params) == 1) & isinstance(list(params.values())[0], GeneratorType):
                if self.verbose:
                    print('found parametric generator')
                # raise NotImplementedError('Animations not yet supported')
                params.update({variable: result})
            else:
                if self.verbose:
                    print('evaluate found generator function. Concatonating solutions')
                    print('generator params:', params)

                params = concat_solution(result, variable)
        else:
            params.update({variable: result})
        return params

    def solve(self, fprime, interval, y0,
              dense_output=True,  # generate a callable solution
              events=None,  # stop when event is triggered
              vectorized=True, ):
        from scipy.integrate import solve_ivp

        result = solve_ivp(self[fprime], interval, y0,
                           dense_output=dense_output,
                           events=events,
                           vectorized=vectorized)
        if self.verbose:
            print(result['message'])

        varname = next(iter(inspect.signature(self[fprime]).parameters.keys()))

        scope = {'result': result, 'varname': varname}

        soln_str = r"""def solution({varname} = result['t']):
            return result['sol'].__call__({varname}).T.squeeze()""".format(varname=varname)

        exec(soln_str.format(), scope)

        return scope['solution']

    def figure(self, variable, indexing='ij', **kwargs):
        """Generates a plotly figure for a given variable and keyword arguments"""
        result = self.evaluate(variable, **kwargs)
        signature = self.signatures[variable]
        units = signature['units']
        if units != '':
            units = '[{}]'.format(units)
        title = self.to_latex([variable], 'inline')
        title_lhs = title.split(' =')[0] + '$'
        title_short = '{}'.format(variable + units)  # something wrong with colorbar latex being vertical
        titles = dict(
            title=title, title_lhs=title_lhs,
            title_short=title_short, units=units, variable=variable)
        fig = dict()
        chart_type = None
        traces = []

        hidden_args = []
        if 'hidden_args' in self[variable].meta:
            hidden_args = self[variable].meta['hidden_args']
        arg_arrays = [result[k] for k in result if k not in hidden_args][:-1]

        arg_shapes = get_arg_shapes(*arg_arrays)

        if isinstance(result[variable], GeneratorType):
            # if evaluate returned a generator, then assume animation
            raise NotImplementedError('Animations not yet supported!')

        try:
            out_dim, *arg_dims = symbolic_shape(result[variable].shape, *arg_shapes)
        except IndexError:
            print('could not interpret shapes from variable shape {}'.format(
                result[variable].shape))
            print('argument shapes: ', *arg_shapes)
            raise

        try:
            plot_func = plot_dict[out_dim][tuple(arg_dims)]['func']
        except KeyError:
            print('not supported: out_dim {}, arg_dims {}'.format(out_dim, arg_dims))
            raise

        traces, chart_type, layout = plot_func(
            result,
            titles,
            indexing=indexing,
            verbose=self.verbose,
            **kwargs)

        layout.update(
            dict(autosize=False,
                 width=700,
                 height=400,
                 margin=go.layout.Margin(
                     # l=30,
                     r=30,
                     b=32,
                     t=40,
                     pad=0),
                 ))

        fig['data'] = traces
        fig['layout'] = layout
        return go.Figure(fig).update_traces(meta=chart_type)
        # if return_type:
        #     fig['chart_type'] = chart_type
        # return fig

    def plot(self, *variables, **figures):
        for k in variables:
            figures[k] = {}
        if len(figures) == 1:
            variable, kwargs = list(figures.items())[0]
            fig = self.figure(variable, **kwargs)
            # if fig['chart_type'] is None:
            #     raise AttributeError("No chart_type for this trace")
            # else:
            #     if self.verbose:
            #         print('chart type:', fig['chart_type'])
            return go.Figure(
                data=fig['data'],
                layout=fig['layout'])
        else:
            traces = []
            layouts = []
            for variable, kwargs in list(figures.items()):
                fig = self.figure(variable, **kwargs)
                traces.extend(fig['data'])
                layouts.append(fig['layout'])
            # Todo: merge the layouts instead of selecting the last one
            return go.Figure(data=traces, layout=layouts[-1])


class KamodoAPI(Kamodo):
    """JSON api wrapper for kamodo services"""
    def __init__(self, url_path, **kwargs):
        self._url_path = url_path
        super(KamodoAPI, self).__init__(**kwargs)

        self._kdata = self._get(self._url_path)

        self._defaults = {}
        self._data = {}

        self.__doc__ = requests.get('{}/doc'.format(self._url_path)).text

        for k, v in self._kdata.items():
            # get defaults for this func
            default_path = '{}/{}/defaults'.format(self._url_path, k)
            self._defaults[k] = self._get(default_path)

            # get cached data (result of calling with no args)
            data_path = '{}/{}/data'.format(self._url_path, k)
            self._data[k] = self._get(data_path)

            doc_path = '{}/{}/doc'.format(self._url_path, k)
            func_doc = requests.get(doc_path).text

            # get kamodofied function
            func = self.load_func(k)
            func.__doc__ += '\n' + func_doc
            self[k] = kamodofy(
                func,
                data=self._data[k],
                units=v['units'],
                )

    def _get(self, url_path):
        req_result = requests.get(url_path)
        try:
            result = req_result.json() # returns a dictionary maybe
        except json.JSONDecodeError as m:
            print('could not decode request {}'.format(req_result.text))
            raise json.JSONDecodeError(m)

        if isinstance(result, str):
            result_dict = json.loads(result, object_hook=deserialize)
        else:
            result_dict = result
        return result_dict

    def _call_func(self, func_name, **kwargs):
        """construct the url and call api with params"""
        url_path = '{}/{}'.format(self._url_path, func_name)
        if self.verbose:
            print('querying {}'.format(url_path))
        params = []
        for k, v in kwargs.items():
            params.append((k, json.dumps(v, default=serialize)))
        result = requests.get(
            url=url_path,
            params=params).json() #returns a dictionary

        if isinstance(result, str):
            if self.verbose:
                print('reloading json as str')
            result = json.loads(result, object_hook=deserialize)
        else:
            result = deserialize(result)
        return result

    def load_func(self, func_name):
        """loads a function signature"""
        signature = []
        for arg, arg_default in self._defaults[func_name].items():
            signature.append(forge.arg(arg, default=arg_default))

        @forge.sign(*signature)
        def api_func(*args, **kwargs):
            """API function"""
            return self._call_func(func_name, **kwargs)

        api_func.__name__ = func_name
        api_func.__doc__ = "{}/{}".format(self._url_path, func_name)
        return api_func

def compose(**kamodos):
    """Kamposes multiple kamodo instances into one"""
    kamodo = Kamodo()
    for kname, k in kamodos.items():
        for name, symbol in k.symbol_registry.items():
            signature = k.signatures[name]
            meta = k[symbol].meta
            data = getattr(k[symbol], 'data', None)

            rhs = signature['rhs']
            registry_name = '{}_{}'.format(name, kname)
            if (rhs is None) | hasattr(rhs, '__call__'):
                kamodo[registry_name] = kamodofy(k[symbol], data=data, **meta)
            else:
                kamodo[registry_name] = str(rhs)

    return kamodo

def from_kamodo(kobj, **funcs):
    """copies a kamodo object, inserting additional functions"""
    knew = Kamodo()
    for name, signature in kobj.signatures.items():
        symbol = signature['symbol']
        knew[symbol] = kobj[symbol]
    for symbol, func in funcs.items():
        knew[symbol] = func
    return knew

def get_figures(func, iterator, verbose=False):
    plots = []
    for a in func(iterator):
        if verbose:
            print('registering {}'.format(a.__name__), end=' ')
        k = Kamodo(**{a.__name__: a})
        if verbose:
            print('getting plot for {}'.format(a.__name__), end=' ')
        fig_plot = k.plot(a.__name__)
        if verbose:
            print('calling full_figure_for_development', end=' ')
        # full_fig = fig_plot.full_figure_for_development(warn=False)
        full_fig = fig_plot
        if verbose:
            print('appending {}'.format(a.__name__))
        plots.append(full_fig)
    return plots

def animate(func_, iterator=None, verbose=False):
    defaults = get_defaults(func_)
    if len(defaults) > 1:
        raise NotImplementedError(
            "Animations with {} args not yet supported".format(len(defaults)))

    if iterator is None:
        param, iterator = list(defaults.items())[0]
    else:
        param = list(defaults.keys())[0]
        iterator = list(iterator)

    figures = get_figures(func_, iterator, verbose)

#     print(len(figures), ' figures')

    axes_ranges = get_ranges(figures)

    layout = figures[0]['layout']
    layout.update(axes_ranges)
    # make figure

    fig_dict = {
        "data": figures[0]['data'],
        "layout": layout,
        "frames": []
    }


    fig_dict["layout"]["updatemenus"] = [
        {
            "buttons": [
                {
                    "args": [None, {"frame": {"duration": 100, "redraw": True},
                                    "fromcurrent": True, "transition": {"duration": 100,
                                                                        "easing": "quadratic-in-out"}}],
                    "label": "Play",
                    "method": "animate",
                },
                {
                    "args": [[None], {"frame": {"duration": 0, "redraw": False},
                                      "mode": "immediate",
                                      "transition": {"duration": 0}}],
                    "label": "Pause",
                    "method": "animate"
                }
            ],
            "direction": "left",
            "pad": {"r": 10, "t": 87},
            "showactive": False,
            "type": "buttons",
            "x": 0.1,
            "xanchor": "right",
            "y": 0,
            "yanchor": "top"
        }
    ]

    sliders_dict = {
        "active": 0,
        "yanchor": "top",
        "xanchor": "left",
        "currentvalue": {
            "font": {"size": 20},
            "prefix": "{}: ".format(param),
            "visible": True,
            "xanchor": "right"
        },
        "transition": {"duration": 100, "easing": "cubic-in-out"},
        "pad": {"b": 10, "t": 50},
        "len": 0.9,
        "x": 0.1,
        "y": 0,
        "steps": []
    }

    # make frames
    for p, figure in zip(iterator, figures):
        frame = {"data": figure['data'],
                 "name": str(p)}
#         print(figure['layout']['xaxis']['range'])

        fig_dict["frames"].append(frame)
        slider_step = {"args": [
            [p],
            {"frame": {"duration": 100, "redraw": True},
             "mode": "immediate",
             "transition": {"duration": 300}}],
            "label": '{:.2f}'.format(p),
            "method": "animate"}
        sliders_dict["steps"].append(slider_step)
#     print(len(fig_dict['frames']), ' frames')

#     fig_dict["data"] = fig_dict["data"] + fig_dict["frames"][0]["data"]
    fig_dict["layout"]["sliders"] = [sliders_dict]

    fig = go.Figure(fig_dict)
    return fig


